#include <stdbool.h>
#include "include/file.h0"
#include "include/parse.h0"
#include "readfile.h"
#include "lib/xalloc.h"
#include "lib/contracts.h"
#include "c0vm.h"
#include <assert.h>
#include <stdlib.h>

int theint(string s) {
  int* pint = parse_int(s,16);
  assert(pint != NULL);
  int res = *pint;
  free(pint);
  return res;
}

void read_error(char *s, size_t i)
{
  fprintf(stderr, "Invalid bytecode at %zu: %s\n", i, s);
  exit(EXIT_FAILURE);
  return;
}

uint32_t read_u32(char** s, size_t i, size_t length)
{
  assert(i + 4 <= length);
  uint32_t sum = 0;
  for (size_t j = i; j < i+4; j++) {
    int x = theint(s[j]);
    if (!(0 <= x && x < 256)) read_error(s[j], j);
    sum = (sum<<8) | theint(s[j]);
  }
  return sum;
}

uint16_t read_u16(char** s, size_t i, size_t length)
{
  assert(i + 2 <= length);
  uint16_t sum = 0;
  for (size_t j = i; j < i+2; j++) {
    int x = theint(s[j]);
    if (!(0 <= x && x < 256)) read_error(s[j], j);
    sum = (sum<<8) | x;
  }
  return sum;

}

ubyte read_u8(char** s, size_t i, size_t length)
{
  assert(i < length);
  int x = theint(s[i]);
  if (!(0 <= x && x < 256)) read_error(s[i], i);
  return (ubyte) x;
}

struct bc0_file* read_program(string filename)
{
  string_bundle sb = read_words(filename);
  char** sb_array = string_bundle_array(sb);
  size_t sb_array_length = string_bundle_length(sb);

  struct bc0_file* bc0 = xmalloc(sizeof(struct bc0_file));
  size_t i = 0;   /* index into bc0_file byte position */

  bc0->magic = read_u32(sb_array, i, sb_array_length);
  i += 4;
  bc0->version = read_u16(sb_array, i, sb_array_length);
  i += 2;

  bc0->int_pool_count = read_u16(sb_array, i, sb_array_length);
  i += 2;

  bc0->int_pool = xcalloc(bc0->int_pool_count, sizeof(int32_t));
  for (size_t j = 0; j < bc0->int_pool_count; j++) {
    bc0->int_pool[j] = read_u32(sb_array, i, sb_array_length);
    i += 4;
  }

  bc0->string_pool_count = read_u16(sb_array, i, sb_array_length);
  i += 2;
  bc0->string_pool = xcalloc(bc0->string_pool_count, sizeof(char));
  for (size_t j = 0; j < bc0->string_pool_count; j++) {
    bc0->string_pool[j] = (char) read_u8(sb_array, i, sb_array_length);
    i++;
  }

  bc0->function_count = read_u16(sb_array, i, sb_array_length);
  i += 2;
  bc0->function_pool = xcalloc(bc0->function_count, sizeof(struct function_info));
  for (size_t j = 0; j < bc0->function_count; j++) {
    bc0->function_pool[j].num_args = read_u16(sb_array, i, sb_array_length);
    i += 2;
    bc0->function_pool[j].num_vars = read_u16(sb_array, i, sb_array_length);
    i += 2;
    bc0->function_pool[j].code_length = read_u16(sb_array, i, sb_array_length);
    i += 2;
    bc0->function_pool[j].code = xcalloc(bc0->function_pool[j].code_length, sizeof(ubyte));
    for (size_t k = 0; k < bc0->function_pool[j].code_length; k++) {
      bc0->function_pool[j].code[k] = read_u8(sb_array, i, sb_array_length);
      i++;
    }
  }

  bc0->native_count = read_u16(sb_array, i, sb_array_length);
  i += 2;
  bc0->native_pool = xcalloc(bc0->native_count, sizeof(struct native_info));
  for (size_t j = 0; j < bc0->native_count; j++) {
    bc0->native_pool[j].num_args = read_u16(sb_array, i, sb_array_length);
    i += 2;
    bc0->native_pool[j].function_table_index = read_u16(sb_array, i, sb_array_length);
    i += 2;
  }

  /* free bundle and all contained strings */
  string_bundle_free(sb, free);

  return bc0;
}

void free_program(struct bc0_file *program)
{
  free(program->int_pool);
  free(program->string_pool);

  for (size_t j = 0; j < program->function_count; j++)
    free(program->function_pool[j].code);
  free(program->function_pool);

  free(program->native_pool);
  free(program);
}
